"""
Module containing the actual `gedcom.parser.Parser` used to generate elements - out of each line -
which can in return be manipulated.
"""

import re as regex
from sys import version_info
from gedcom.element.element import Element
from gedcom.element.family import FamilyElement, NotAnActualFamilyError
from gedcom.element.file import FileElement
from gedcom.element.individual import IndividualElement, NotAnActualIndividualError
from gedcom.element.object import ObjectElement
from gedcom.element.root import RootElement
import gedcom.tags

FAMILY_MEMBERS_TYPE_ALL = "ALL"
FAMILY_MEMBERS_TYPE_CHILDREN = gedcom.tags.GEDCOM_TAG_CHILD
FAMILY_MEMBERS_TYPE_HUSBAND = gedcom.tags.GEDCOM_TAG_HUSBAND
FAMILY_MEMBERS_TYPE_PARENTS = "PARENTS"
FAMILY_MEMBERS_TYPE_WIFE = gedcom.tags.GEDCOM_TAG_WIFE


class GedcomFormatViolationError(Exception):
    pass


class PointerNotFoundException(Exception):
    pass


class Parser(object):
    """Parses and manipulates GEDCOM 5.5 format data

    For documentation of the GEDCOM 5.5 format, see: https://homepages.rootsweb.com/~pmcbride/gedcom/55gctoc.htm

    This parser reads and parses a GEDCOM file.

    Elements may be accessed via:

    * a `list` through `gedcom.parser.Parser.get_element_list()`
    * a `dict` through `gedcom.parser.Parser.get_element_dictionary()`
    """

    def __init__(self):
        self.__element_list = []
        self.__element_dictionary = {}
        self.__root_element = RootElement()

    def invalidate_cache(self):
        """Empties the element list and dictionary to cause `gedcom.parser.Parser.get_element_list()`
        and `gedcom.parser.Parser.get_element_dictionary()` to return updated data.

        The update gets deferred until each of the methods actually gets called.
        """
        self.__element_list = []
        self.__element_dictionary = {}

    def get_element_by_pointer(self, pointer):
        """Returns the element that has the provided pointer. Raises an exception if that pointer doesn't exist.
        :type pointer: string
        :rtype: Element
        """
        element_dictionary = self.get_element_dictionary()
        if pointer not in element_dictionary:
            raise PointerNotFoundException("No element with the pointer " + pointer + " was found.")
        else:
            return element_dictionary[pointer]

    def get_element_list(self):
        """Returns a list containing all elements from within the GEDCOM file

        By default elements are in the same order as they appeared in the file.

        This list gets generated on-the-fly, but gets cached. If the database
        was modified, you should call `gedcom.parser.Parser.invalidate_cache()` once to let this
        method return updated data.

        Consider using `gedcom.parser.Parser.get_root_element()` or `gedcom.parser.Parser.get_root_child_elements()` to access
        the hierarchical GEDCOM tree, unless you rarely modify the database.

        :rtype: list of Element
        """
        if not self.__element_list:
            for element in self.get_root_child_elements():
                self.__build_list(element, self.__element_list)
        return self.__element_list

    def get_element_dictionary(self):
        """Returns a dictionary containing all elements, identified by a pointer, from within the GEDCOM file

        Only elements identified by a pointer are listed in the dictionary.
        The keys for the dictionary are the pointers.

        This dictionary gets generated on-the-fly, but gets cached. If the
        database was modified, you should call `invalidate_cache()` once to let
        this method return updated data.

        :rtype: dict[str, Element]
        """
        if not self.__element_dictionary:
            self.__element_dictionary = {
                element.get_pointer(): element for element in self.get_root_child_elements() if element.get_pointer()
            }

        return self.__element_dictionary

    def get_root_element(self):
        """Returns a virtual root element containing all logical records as children

        When printed, this element converts to an empty string.

        :rtype: RootElement
        """
        return self.__root_element

    def get_root_child_elements(self):
        """Returns a list of logical records in the GEDCOM file

        By default, elements are in the same order as they appeared in the file.

        :rtype: list of Element
        """
        return self.get_root_element().get_child_elements()

    def parse_file(self, file_path, strict=True):
        """Opens and parses a file, from the given file path, as GEDCOM 5.5 formatted data
        :type file_path: str
        :type strict: bool
        """
        with open(file_path, 'rb') as gedcom_stream:
            self.parse(gedcom_stream, strict)

    def parse(self, gedcom_stream, strict=True):
        """Parses a stream, or an array of lines, as GEDCOM 5.5 formatted data
        :type gedcom_stream: a file stream, or str array of lines with new line at the end
        :type strict: bool
        """
        self.invalidate_cache()
        self.__root_element = RootElement()

        line_number = 1
        last_element = self.get_root_element()

        for line in gedcom_stream:
            last_element = self.__parse_line(line_number, line.decode('utf-8-sig'), last_element, strict)
            line_number += 1

    # Private methods

    @staticmethod
    def __parse_line(line_number, line, last_element, strict=True):
        """Parse a line from a GEDCOM 5.5 formatted document

        Each line should have the following (bracketed items optional):
        level + ' ' + [pointer + ' ' +] tag + [' ' + line_value]

        :type line_number: int
        :type line: str
        :type last_element: Element
        :type strict: bool

        :rtype: Element
        """

        # Level must start with non-negative int, no leading zeros.
        level_regex = '^(0|[1-9]+[0-9]*) '

        # Pointer optional, if it exists it must be flanked by `@`
        pointer_regex = '(@[^@]+@ |)'

        # Tag must be an alphanumeric string
        tag_regex = '([A-Za-z0-9_]+)'

        # Value optional, consists of anything after a space to end of line
        value_regex = '( [^\n\r]*|)'

        # End of line defined by `\n` or `\r`
        end_of_line_regex = '([\r\n]{1,2})'

        # Complete regex
        gedcom_line_regex = level_regex + pointer_regex + tag_regex + value_regex + end_of_line_regex
        regex_match = regex.match(gedcom_line_regex, line)

        if regex_match is None:
            if strict:
                error_message = ("Line <%d:%s> of document violates GEDCOM format 5.5" % (line_number, line)
                                 + "\nSee: https://chronoplexsoftware.com/gedcomvalidator/gedcom/gedcom-5.5.pdf")
                raise GedcomFormatViolationError(error_message)
            else:
                # Quirk check - see if this is a line without a CRLF (which could be the last line)
                last_line_regex = level_regex + pointer_regex + tag_regex + value_regex
                regex_match = regex.match(last_line_regex, line)
                if regex_match is not None:
                    line_parts = regex_match.groups()

                    level = int(line_parts[0])
                    pointer = line_parts[1].rstrip(' ')
                    tag = line_parts[2]
                    value = line_parts[3].strip()
                    crlf = '\n'
                else:
                    # Quirk check - Sometimes a gedcom has a text field with a CR.
                    # This creates a line without the standard level and pointer.
                    # If this is detected then turn it into a CONC or CONT.
                    line_regex = '([^\n\r]*|)'
                    cont_line_regex = line_regex + end_of_line_regex
                    regex_match = regex.match(cont_line_regex, line)
                    line_parts = regex_match.groups()
                    level = last_element.get_level()
                    tag = last_element.get_tag()
                    pointer = None
                    value = line_parts[0].strip()
                    crlf = line_parts[1]
                    if tag != gedcom.tags.GEDCOM_TAG_CONTINUED and tag != gedcom.tags.GEDCOM_TAG_CONCATENATION:
                        # Increment level and change this line to a CONC
                        level += 1
                        tag = gedcom.tags.GEDCOM_TAG_CONCATENATION
        else:
            line_parts = regex_match.groups()

            level = int(line_parts[0])
            pointer = line_parts[1].rstrip(' ')
            tag = line_parts[2]
            value = line_parts[3].strip()
            crlf = line_parts[4]

        # Check level: should never be more than one higher than previous line.
        if level > last_element.get_level() + 1:
            error_message = ("Line %d of document violates GEDCOM format 5.5" % line_number
                             + "\nLines must be no more than one level higher than previous line."
                             + "\nSee: https://chronoplexsoftware.com/gedcomvalidator/gedcom/gedcom-5.5.pdf")
            raise GedcomFormatViolationError(error_message)

        # Create element. Store in list and dict, create children and parents.
        if tag == gedcom.tags.GEDCOM_TAG_INDIVIDUAL:
            element = IndividualElement(level, pointer, tag, value, crlf, multi_line=False)
        elif tag == gedcom.tags.GEDCOM_TAG_FAMILY:
            element = FamilyElement(level, pointer, tag, value, crlf, multi_line=False)
        elif tag == gedcom.tags.GEDCOM_TAG_FILE:
            element = FileElement(level, pointer, tag, value, crlf, multi_line=False)
        elif tag == gedcom.tags.GEDCOM_TAG_OBJECT:
            element = ObjectElement(level, pointer, tag, value, crlf, multi_line=False)
        else:
            element = Element(level, pointer, tag, value, crlf, multi_line=False)

        # Start with last element as parent, back up if necessary.
        parent_element = last_element

        while parent_element.get_level() > level - 1:
            parent_element = parent_element.get_parent_element()

        # Add child to parent & parent to child.
        parent_element.add_child_element(element)

        return element

    def __build_list(self, element, element_list):
        """Recursively add elements to a list containing elements
        :type element: Element
        :type element_list: list of Element
        """
        element_list.append(element)
        for child in element.get_child_elements():
            self.__build_list(child, element_list)

    # Methods for analyzing individuals and relationships between individuals

    def get_marriages(self, individual):
        """Returns a list of marriages of an individual formatted as a tuple (`str` date, `str` place)
        :type individual: IndividualElement
        :rtype: tuple
        """
        marriages = []
        if not isinstance(individual, IndividualElement):
            raise NotAnActualIndividualError(
                "Operation only valid for elements with %s tag" % gedcom.tags.GEDCOM_TAG_INDIVIDUAL
            )
        # Get and analyze families where individual is spouse.
        families = self.get_families(individual, gedcom.tags.GEDCOM_TAG_FAMILY_SPOUSE)
        for family in families:
            for family_data in family.get_child_elements():
                if family_data.get_tag() == gedcom.tags.GEDCOM_TAG_MARRIAGE:
                    date = ''
                    place = ''
                    for marriage_data in family_data.get_child_elements():
                        if marriage_data.get_tag() == gedcom.tags.GEDCOM_TAG_DATE:
                            date = marriage_data.get_value()
                        if marriage_data.get_tag() == gedcom.tags.GEDCOM_TAG_PLACE:
                            place = marriage_data.get_value()
                    marriages.append((date, place))
        return marriages

    def get_marriage_years(self, individual):
        """Returns a list of marriage years (as integers) for an individual
        :type individual: IndividualElement
        :rtype: list of int
        """
        dates = []

        if not isinstance(individual, IndividualElement):
            raise NotAnActualIndividualError(
                "Operation only valid for elements with %s tag" % gedcom.tags.GEDCOM_TAG_INDIVIDUAL
            )

        # Get and analyze families where individual is spouse.
        families = self.get_families(individual, gedcom.tags.GEDCOM_TAG_FAMILY_SPOUSE)
        for family in families:
            for child in family.get_child_elements():
                if child.get_tag() == gedcom.tags.GEDCOM_TAG_MARRIAGE:
                    for childOfChild in child.get_child_elements():
                        if childOfChild.get_tag() == gedcom.tags.GEDCOM_TAG_DATE:
                            date = childOfChild.get_value().split()[-1]
                            try:
                                dates.append(int(date))
                            except ValueError:
                                pass
        return dates

    def marriage_year_match(self, individual, year):
        """Checks if one of the marriage years of an individual matches the supplied year. Year is an integer.
        :type individual: IndividualElement
        :type year: int
        :rtype: bool
        """
        if not isinstance(individual, IndividualElement):
            raise NotAnActualIndividualError(
                "Operation only valid for elements with %s tag" % gedcom.tags.GEDCOM_TAG_INDIVIDUAL
            )

        years = self.get_marriage_years(individual)
        return year in years

    def marriage_range_match(self, individual, from_year, to_year):
        """Check if one of the marriage years of an individual is in a given range. Years are integers.
        :type individual: IndividualElement
        :type from_year: int
        :type to_year: int
        :rtype: bool
        """
        if not isinstance(individual, IndividualElement):
            raise NotAnActualIndividualError(
                "Operation only valid for elements with %s tag" % gedcom.tags.GEDCOM_TAG_INDIVIDUAL
            )

        years = self.get_marriage_years(individual)
        for year in years:
            if from_year <= year <= to_year:
                return True
        return False

    def get_families(self, individual, family_type=gedcom.tags.GEDCOM_TAG_FAMILY_SPOUSE):
        """Return family elements listed for an individual

        family_type can be `gedcom.tags.GEDCOM_TAG_FAMILY_SPOUSE` (families where the individual is a spouse) or
        `gedcom.tags.GEDCOM_TAG_FAMILY_CHILD` (families where the individual is a child). If a value is not
        provided, `gedcom.tags.GEDCOM_TAG_FAMILY_SPOUSE` is default value.

        :type individual: IndividualElement
        :type family_type: str
        :rtype: list of FamilyElement
        """
        if not isinstance(individual, IndividualElement):
            raise NotAnActualIndividualError(
                "Operation only valid for elements with %s tag" % gedcom.tags.GEDCOM_TAG_INDIVIDUAL
            )

        families = []
        element_dictionary = self.get_element_dictionary()

        for child_element in individual.get_child_elements():
            is_family = (child_element.get_tag() == family_type
                         and child_element.get_value() in element_dictionary)
            if is_family:
                families.append(element_dictionary[child_element.get_value()])

        return families

    def get_ancestors(self, individual, ancestor_type="ALL"):
        """Return elements corresponding to ancestors of an individual

        Optional `ancestor_type`. Default "ALL" returns all ancestors, "NAT" can be
        used to specify only natural (genetic) ancestors.

        :type individual: IndividualElement
        :type ancestor_type: str
        :rtype: list of Element
        """
        if not isinstance(individual, IndividualElement):
            raise NotAnActualIndividualError(
                "Operation only valid for elements with %s tag" % gedcom.tags.GEDCOM_TAG_INDIVIDUAL
            )

        parents = self.get_parents(individual, ancestor_type)
        ancestors = []
        ancestors.extend(parents)

        for parent in parents:
            ancestors.extend(self.get_ancestors(parent, ancestor_type))

        return ancestors

    def get_parents(self, individual, parent_type="ALL"):
        """Return elements corresponding to parents of an individual

        Optional parent_type. Default "ALL" returns all parents. "NAT" can be
        used to specify only natural (genetic) parents.

        :type individual: IndividualElement
        :type parent_type: str
        :rtype: list of IndividualElement
        """
        if not isinstance(individual, IndividualElement):
            raise NotAnActualIndividualError(
                "Operation only valid for elements with %s tag" % gedcom.tags.GEDCOM_TAG_INDIVIDUAL
            )

        parents = []
        families = self.get_families(individual, gedcom.tags.GEDCOM_TAG_FAMILY_CHILD)

        for family in families:
            if parent_type == "NAT":
                for family_member in family.get_child_elements():

                    if family_member.get_tag() == gedcom.tags.GEDCOM_TAG_CHILD \
                       and family_member.get_value() == individual.get_pointer():

                        for child in family_member.get_child_elements():
                            if child.get_value() == "Natural":
                                if child.get_tag() == gedcom.tags.GEDCOM_PROGRAM_DEFINED_TAG_MREL:
                                    parents += self.get_family_members(family, gedcom.tags.GEDCOM_TAG_WIFE)
                                elif child.get_tag() == gedcom.tags.GEDCOM_PROGRAM_DEFINED_TAG_FREL:
                                    parents += self.get_family_members(family, gedcom.tags.GEDCOM_TAG_HUSBAND)
            else:
                parents += self.get_family_members(family, "PARENTS")

        return parents

    def find_path_to_ancestor(self, descendant, ancestor, path=None):
        """Return path from descendant to ancestor
        :rtype: object
        """
        if not isinstance(descendant, IndividualElement) or not isinstance(ancestor, IndividualElement):
            raise NotAnActualIndividualError(
                "Operation only valid for elements with %s tag." % gedcom.tags.GEDCOM_TAG_INDIVIDUAL
            )

        if not path:
            path = [descendant]

        if path[-1].get_pointer() == ancestor.get_pointer():
            return path
        else:
            parents = self.get_parents(descendant, "NAT")
            for parent in parents:
                potential_path = self.find_path_to_ancestor(parent, ancestor, path + [parent])
                if potential_path is not None:
                    return potential_path

        return None

    def get_family_members(self, family, members_type=FAMILY_MEMBERS_TYPE_ALL):
        """Return array of family members: individual, spouse, and children

        Optional argument `members_type` can be used to return specific subsets:

        "FAMILY_MEMBERS_TYPE_ALL": Default, return all members of the family
        "FAMILY_MEMBERS_TYPE_PARENTS": Return individuals with "HUSB" and "WIFE" tags (parents)
        "FAMILY_MEMBERS_TYPE_HUSBAND": Return individuals with "HUSB" tags (father)
        "FAMILY_MEMBERS_TYPE_WIFE": Return individuals with "WIFE" tags (mother)
        "FAMILY_MEMBERS_TYPE_CHILDREN": Return individuals with "CHIL" tags (children)

        :type family: FamilyElement
        :type members_type: str
        :rtype: list of IndividualElement
        """
        if not isinstance(family, FamilyElement):
            raise NotAnActualFamilyError(
                "Operation only valid for element with %s tag." % gedcom.tags.GEDCOM_TAG_FAMILY
            )

        family_members = []
        element_dictionary = self.get_element_dictionary()

        for child_element in family.get_child_elements():
            # Default is ALL
            is_family = (child_element.get_tag() == gedcom.tags.GEDCOM_TAG_HUSBAND
                         or child_element.get_tag() == gedcom.tags.GEDCOM_TAG_WIFE
                         or child_element.get_tag() == gedcom.tags.GEDCOM_TAG_CHILD)

            if members_type == FAMILY_MEMBERS_TYPE_PARENTS:
                is_family = (child_element.get_tag() == gedcom.tags.GEDCOM_TAG_HUSBAND
                             or child_element.get_tag() == gedcom.tags.GEDCOM_TAG_WIFE)
            elif members_type == FAMILY_MEMBERS_TYPE_HUSBAND:
                is_family = child_element.get_tag() == gedcom.tags.GEDCOM_TAG_HUSBAND
            elif members_type == FAMILY_MEMBERS_TYPE_WIFE:
                is_family = child_element.get_tag() == gedcom.tags.GEDCOM_TAG_WIFE
            elif members_type == FAMILY_MEMBERS_TYPE_CHILDREN:
                is_family = child_element.get_tag() == gedcom.tags.GEDCOM_TAG_CHILD

            if is_family and child_element.get_value() in element_dictionary:
                family_members.append(element_dictionary[child_element.get_value()])

        return family_members

    # Other methods

    def print_gedcom(self):
        """Write GEDCOM data to stdout"""
        from sys import stdout
        self.save_gedcom(stdout)

    def save_gedcom(self, open_file):
        """Save GEDCOM data to a file
        :type open_file: file
        """
        if version_info[0] >= 3:
            open_file.write(self.get_root_element().to_gedcom_string(True))
        else:
            open_file.write(self.get_root_element().to_gedcom_string(True).encode('utf-8-sig'))



import unittest

from gedcom.parser import Parser

from gedcom.element.individual import IndividualElement


class TestIndividualElement(unittest.TestCase):

    def test_is_deceased__a_person_without_a_death_element_should_not_be_deceased(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 SEX M
                1 BIRT
                    2 DATE 1 JAN 1900
                1 NAME Second /Surname/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertFalse(individual.is_deceased())

    def test_is_deceased__a_person_with_a_death_element_should_be_deceased(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 SEX M
                1 DEAT
                    2 DATE 1 JAN 1900
                1 NAME Second /Surname/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertTrue(individual.is_deceased())

    # --------------------- START OF is_child TESTING -----------------------

    def test_is_child__a_person_not_in_a_family_is_not_a_child(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertFalse(individual.is_child())

    def test_is_child__a_person_in_a_family_but_only_as_a_parent_is_not_a_child(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 FAMS @F1@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertFalse(individual.is_child())

    def test_is_child__a_person_in_a_family_and_as_a_child_is_a_child(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 FAMC @F1@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertTrue(individual.is_child())

    # --------------------- START OF is_spouse TESTING -----------------------

    def test_is_spouse__a_person_not_in_a_family_is_not_a_spouse(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertFalse(individual.is_spouse())

    def test_is_spouse__a_person_in_a_family_but_only_as_a_child_is_not_a_spouse(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 FAMC @F1@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertFalse(individual.is_spouse())

    def test_is_spouse__a_person_in_a_family_and_as_a_spouse_is_a_spouse(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 FAMS @F1@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertTrue(individual.is_spouse())

    # --------------------- START OF is_private TESTING -----------------------

    def test_is_private__a_person_not_marked_private_is_not_private(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertFalse(individual.is_private())

    def test_is_child__a_person_with_a_private_tag_but_not_marked_yes_is_not_private(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 PRIV N
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertFalse(individual.is_private())

    def test_is_child__a_person_with_a_private_tag_and_marked_yes_is_private(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 PRIV Y
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertTrue(individual.is_private())

    # --------------------- START OF get_name TESTING -----------------------

    def test_get_name__should_return_full_name_of_a_person_with_a_name_tag_that_has_a_value(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(('First', 'Last'), individual.get_name())

    def test_get_name__should_remove_suffix_of_a_persons_name_when_they_have_one(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/ Jr
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(('First', 'Last'), individual.get_name())

    def test_get_name__should_return_full_name_of_a_person_with_a_name_tag_that_has_sub_elements_with_values(self):
        use_case = """
            0 @I1@ INDI
                1 NAME
                    2 GIVN First
                    2 SURN Last
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(('First', 'Last'), individual.get_name())

    def test_get_name__should_return_full_name_of_a_person_with_a_name_tag_that_only_has_given_name_sub_element(self):
        use_case = """
            0 @I1@ INDI
                1 NAME
                    2 GIVN First
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(('First', ''), individual.get_name())

    def test_get_name__should_return_full_name_of_a_person_with_a_name_tag_that_only_has_surname_sub_element(self):
        use_case = """
            0 @I1@ INDI
                1 NAME
                    2 SURN Last
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(('', 'Last'), individual.get_name())

    def test_get_name__should_return_given_name_of_a_person_with_a_name_tag_that_only_has_a_given_name(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(('First', ''), individual.get_name())

    def test_get_name__should_return_surname_name_of_a_person_with_a_name_tag_that_only_has_a_surname(self):
        use_case = """
            0 @I1@ INDI
                1 NAME /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(('', 'Last'), individual.get_name())

    def test_get_name__should_return_empty_strings_for_a_person_without_a_name_tag(self):
        use_case = """
            0 @I1@ INDI
                1 BIRT
                    2 DATE Jan 1 1900
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(('', ''), individual.get_name())

    # --------------------- START OF get_all_names TESTING -----------------------

    def test_get_all_names__should_be_able_to_match_multiple_names_but_not_non_names(self):
        element = IndividualElement(level=0, pointer="@I5@", tag="INDI", value="")
        element.new_child_element(tag="NAME", value="First /Last/")
        element.new_child_element(tag="SEX", value="M")
        birth = element.new_child_element(tag="BIRT", value="")
        birth.new_child_element(tag="DATE", value="1 JAN 1900")
        element.new_child_element(tag="NAME", value="Second /Surname/")

        all_names = element.get_all_names()
        self.assertEqual(2, len(all_names))
        self.assertEqual('First /Last/', all_names[0])
        self.assertEqual('Second /Surname/', all_names[1])

    # --------------------- START OF surname_match TESTING -----------------------

    def test_surname_match__should_match_exactly_a_surname(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertTrue(individual.surname_match('Last'))

    def test_surname_match__should_match_a_surname_while_ignoring_case(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertTrue(individual.surname_match('LAST'))

    def test_surname_match__should_not_match_a_surname_if_different(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertFalse(individual.surname_match('not a match'))

    # --------------------- START OF given_name_match TESTING -----------------------

    def test_given_name_match__should_match_exactly_a_given_name(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertTrue(individual.given_name_match('First'))

    def test_given_name_match__should_match_a_given_name_while_ignoring_case(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertTrue(individual.given_name_match('FIRST'))

    def test_given_name_match__should_not_match_a_given_name_if_different(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertFalse(individual.given_name_match('not a match'))

    # --------------------- START OF get_gender TESTING -----------------------

    def test_get_gender__should_get_the_persons_gender_if_present(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 SEX M
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual("M", individual.get_gender())

    def test_get_gender__should_return_an_empty_string_if_gender_is_not_present(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual("", individual.get_gender())

    # --------------------- START OF get_birth_data TESTING -----------------------

    def test_get_birth_data__should_return_an_empty_tuple_if_no_such_data_is_present(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 SEX M
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("", "", []), individual.get_birth_data())

    def test_get_birth_data__should_return_a_single_tuple_when_there_is_just_a_date(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BIRT
                    2 DATE JUN 1990
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("JUN 1990", "", []), individual.get_birth_data())

    def test_get_birth_data__should_return_a_single_tuple_when_there_is_just_a_place(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BIRT
                    2 PLAC Chicago
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("", "Chicago", []), individual.get_birth_data())

    def test_get_birth_data__should_return_a_single_tuple_when_there_is_just_a_source(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BIRT
                    2 SOUR @S1@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("", "", ["@S1@"]), individual.get_birth_data())

    def test_get_birth_data__should_return_a_single_full_tuple_when_there_is_a_date_place_and_multiple_sources(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BIRT
                    2 DATE APR 2001
                    2 PLAC New York
                    2 SOUR @S1@
                    2 SOUR @S2@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("APR 2001", "New York", ["@S1@", "@S2@"]), individual.get_birth_data())

    def test_get_birth_data__should_return_a_combination_of_fields_in_a_tuple_when_there_are_multiple_partials(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BIRT
                    2 DATE APR 2001
                1 BIRT
                    2 PLAC New York
                1 BIRT
                    2 SOUR @S1@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("APR 2001", "New York", ["@S1@"]), individual.get_birth_data())

    # --------------------- START OF get_death_data TESTING -----------------------

    def test_get_death_data__should_return_an_empty_tuple_if_no_such_data_is_present(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 SEX M
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("", "", []), individual.get_death_data())

    def test_get_death_data__should_return_a_single_tuple_when_there_is_just_a_date(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 DEAT
                    2 DATE JUN 1990
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("JUN 1990", "", []), individual.get_death_data())

    def test_get_death_data__should_return_a_single_tuple_when_there_is_just_a_place(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 DEAT
                    2 PLAC Chicago
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("", "Chicago", []), individual.get_death_data())

    def test_get_death_data__should_return_a_single_tuple_when_there_is_just_a_source(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 DEAT
                    2 SOUR @S1@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("", "", ["@S1@"]), individual.get_death_data())

    def test_get_death_data__should_return_a_single_full_tuple_when_there_is_a_date_place_and_multiple_sources(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 DEAT
                    2 DATE APR 2001
                    2 PLAC New York
                    2 SOUR @S1@
                    2 SOUR @S2@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("APR 2001", "New York", ["@S1@", "@S2@"]), individual.get_death_data())

    def test_get_death_data__should_return_a_combination_of_fields_in_a_tuple_when_there_are_multiple_partials(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 DEAT
                    2 DATE APR 2001
                1 DEAT
                    2 PLAC New York
                1 DEAT
                    2 SOUR @S1@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("APR 2001", "New York", ["@S1@"]), individual.get_death_data())

    # --------------------- START OF get_birth_year TESTING -----------------------

    def test_get_birth_year__should_remove_about_string_from_a_year(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BIRT
                    2 DATE ABT 1924
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(1924, individual.get_birth_year())

    def test_get_birth_year__should_remove_after_string_from_a_year(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BIRT
                    2 DATE AFT 1924
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(1924, individual.get_birth_year())

    def test_get_birth_year__should_remove_before_string_from_a_year(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BIRT
                    2 DATE AFT 1924
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(1924, individual.get_birth_year())

    def test_get_birth_year__should_handle_case_where_there_is_no_year(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BIRT
                    2 DATE JUN
                    2 PLAC Chicago
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(-1, individual.get_birth_year())

    def test_get_birth_year__should_return_birth_year_not_found_value_if_no_date_tag_exists(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(-1, individual.get_birth_year())

    # --------------------- START OF get_death_year TESTING -----------------------

    def test_get_death_year__should_remove_about_string_from_a_year(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 DEAT
                    2 DATE ABT 1924
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(1924, individual.get_death_year())

    def test_get_death_year__should_remove_after_string_from_a_year(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 DEAT
                    2 DATE AFT 1924
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(1924, individual.get_death_year())

    def test_get_death_year__should_remove_before_string_from_a_year(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 DEAT
                    2 DATE AFT 1924
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(1924, individual.get_death_year())

    def test_get_death_year__should_handle_case_where_there_is_no_year(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 DEAT
                    2 DATE JUN
                    2 PLAC Chicago
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(-1, individual.get_death_year())

    def test_get_death_year__should_return_death_year_not_found_value_if_no_date_tag_exists(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(-1, individual.get_death_year())

    # --------------------- START OF get_burial_data TESTING -----------------------

    def test_get_burial_data__should_return_an_empty_tuple_if_no_such_data_is_present(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 SEX M
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("", "", []), individual.get_burial_data())

    def test_get_burial_data__should_return_a_single_tuple_when_there_is_just_a_date(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BURI
                    2 DATE JUN 1990
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("JUN 1990", "", []), individual.get_burial_data())

    def test_get_burial_data__should_return_a_single_tuple_when_there_is_just_a_place(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BURI
                    2 PLAC Chicago
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("", "Chicago", []), individual.get_burial_data())

    def test_get_burial_data__should_return_a_single_tuple_when_there_is_just_a_source(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BURI
                    2 SOUR @S1@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("", "", ["@S1@"]), individual.get_burial_data())

    def test_get_burial_data__should_return_a_single_full_tuple_when_there_is_a_date_place_and_multiple_sources(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BURI
                    2 DATE APR 2001
                    2 PLAC New York
                    2 SOUR @S1@
                    2 SOUR @S2@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("APR 2001", "New York", ["@S1@", "@S2@"]), individual.get_burial_data())

    def test_get_burial_data__should_return_a_combination_of_fields_in_a_tuple_when_there_are_multiple_partials(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 BURI
                    2 DATE APR 2001
                1 BURI
                    2 PLAC New York
                1 BURI
                    2 SOUR @S1@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual(("APR 2001", "New York", ["@S1@"]), individual.get_burial_data())

    # --------------------- START OF get_census_data TESTING -----------------------

    def test_get_census_data__should_return_an_empty_list_if_no_such_data_is_present(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 SEX M
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual([], individual.get_census_data())

    def test_get_census_data__should_return_a_list_containing_a_single_tuple_when_there_is_just_a_date(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 CENS
                    2 DATE JUN 1990
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual([("JUN 1990", "", [])], individual.get_census_data())

    def test_get_census_data__should_return_a_list_containing_a_single_tuple_when_there_is_just_a_place(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 CENS
                    2 PLAC Chicago
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual([("", "Chicago", [])], individual.get_census_data())

    def test_get_census_data__should_return_a_list_containing_a_single_tuple_when_there_is_just_a_source(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 CENS
                    2 SOUR @S1@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual([("", "", ["@S1@"])], individual.get_census_data())

    def test_get_census_data__should_return_a_list_containing_a_single_full_tuple_when_there_is_a_date_place_and_multiple_sources(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 CENS
                    2 DATE APR 2001
                    2 PLAC New York
                    2 SOUR @S1@
                    2 SOUR @S2@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual([("APR 2001", "New York", ["@S1@", "@S2@"])], individual.get_census_data())

    def test_get_census_data__should_return_separate_tuples_when_there_are_multiple_partials(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 CENS
                    2 DATE APR 2001
                1 CENS
                    2 PLAC New York
                1 CENS
                    2 SOUR @S1@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual([("APR 2001", "", []), ("", "New York", []), ("", "", ["@S1@"])], individual.get_census_data())

    # --------------------- START OF get_last_change_date TESTING -----------------------

    def test_get_last_change_date__should_return_empty_string_if_there_is_no_last_change_date(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual("", individual.get_last_change_date())

    def test_get_last_change_date__should_return_the_date_if_there_is_a_last_change_date(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 CHAN
                    2 DATE Jun 1990
                    2 SOUR @S1@
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual("Jun 1990", individual.get_last_change_date())

    # --------------------- START OF get_occupation TESTING -----------------------

    def test_get_occupation__should_return_empty_string_if_there_is_no_occupation(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual("", individual.get_occupation())

    def test_get_occupation__should_return_the_occupation_if_there_is_one(self):
        use_case = """
            0 @I1@ INDI
                1 NAME First /Last/
                1 OCCU Plumber
                1 SEX M
        """
        individual = self._parse_use_case_and_get_individual_element(use_case, "@I1@")
        self.assertEqual("Plumber", individual.get_occupation())

    # --------------------- START OF birth_year_match TESTING -----------------------

    _birth_year_match_use_case = """
        0 @I1@ INDI
            1 NAME First /Last/
            1 BIRT
                2 DATE April 1 1990
    """

    def test_birth_year_match__should_return_true_if_the_year_matches(self):
        individual = self._parse_use_case_and_get_individual_element(self._birth_year_match_use_case, "@I1@")
        self.assertTrue(individual.birth_year_match(1990))

    def test_birth_year_match__should_return_false_if_the_year_does_not_match(self):
        individual = self._parse_use_case_and_get_individual_element(self._birth_year_match_use_case, "@I1@")
        self.assertFalse(individual.birth_year_match(1800))

    # --------------------- START OF birth_range_match TESTING -----------------------

    _birth_range_match_use_case = """
        0 @I1@ INDI
            1 NAME First /Last/
            1 BIRT
                2 DATE April 1 1990
    """

    def test_birth_year_match__should_return_false_if_the_year_is_before_the_range(self):
        individual = self._parse_use_case_and_get_individual_element(self._birth_range_match_use_case, "@I1@")
        self.assertFalse(individual.birth_range_match(2000, 2100))

    def test_birth_year_match__should_return_true_if_the_year_is_in_the_range(self):
        individual = self._parse_use_case_and_get_individual_element(self._birth_range_match_use_case, "@I1@")
        self.assertTrue(individual.birth_range_match(1900, 2000))

    def test_birth_year_match__should_return_false_if_the_year_is_after_the_range(self):
        individual = self._parse_use_case_and_get_individual_element(self._birth_range_match_use_case, "@I1@")
        self.assertFalse(individual.birth_range_match(1800, 1900))

    # --------------------- START OF death_year_match TESTING -----------------------

    _death_year_match_use_case = """
        0 @I1@ INDI
            1 NAME First /Last/
            1 DEAT
                2 DATE April 1 1990
    """

    def test_death_year_match__should_return_true_if_the_year_matches(self):
        individual = self._parse_use_case_and_get_individual_element(self._death_year_match_use_case, "@I1@")
        self.assertTrue(individual.death_year_match(1990))

    def test_death_year_match__should_return_false_if_the_year_does_not_match(self):
        individual = self._parse_use_case_and_get_individual_element(self._death_year_match_use_case, "@I1@")
        self.assertFalse(individual.death_year_match(1800))

    # --------------------- START OF death_range_match TESTING -----------------------

    _death_range_match_use_case = """
        0 @I1@ INDI
            1 NAME First /Last/
            1 DEAT
                2 DATE April 1 1990
    """

    def test_death_year_match__should_return_false_if_the_year_is_before_the_range(self):
        individual = self._parse_use_case_and_get_individual_element(self._death_range_match_use_case, "@I1@")
        self.assertFalse(individual.death_range_match(2000, 2100))

    def test_death_year_match__should_return_true_if_the_year_is_in_the_range(self):
        individual = self._parse_use_case_and_get_individual_element(self._death_range_match_use_case, "@I1@")
        self.assertTrue(individual.death_range_match(1900, 2000))

    def test_death_year_match__should_return_false_if_the_year_is_after_the_range(self):
        individual = self._parse_use_case_and_get_individual_element(self._death_range_match_use_case, "@I1@")
        self.assertFalse(individual.death_range_match(1800, 1900))

    # --------------------- START OF criteria_match TESTING -----------------------

    _criteria_match_use_case = """
        0 @I1@ INDI
            1 NAME First /Last/
            1 BIRT
                2 DATE April 1 1990
            1 DEAT
                2 DATE Jun 21 1999
    """

    def test_criteria_match__should_match_on_matching_surname(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertTrue(individual.criteria_match("surname=Last"))

    def test_criteria_match__should_fail_to_match_on_matching_surname(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("surname=NotAMatch"))

    def test_criteria_match__should_match_on_matching_given_name(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertTrue(individual.criteria_match("name=First"))

    def test_criteria_match__should_fail_to_match_on_matching_given_name(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("name=NotAMatch"))

    def test_criteria_match__should_match_on_matching_birth_year(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertTrue(individual.criteria_match("birth=1990"))

    def test_criteria_match__should_fail_to_match_on_matching_birth_year(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("birth=2127"))

    def test_criteria_match__should_fail_to_match_when_the_provided_birth_year_is_not_numeric(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("birth=not_a_number"))

    def test_criteria_match__should_match_on_matching_birth_range(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertTrue(individual.criteria_match("birth_range=1900-2000"))

    def test_criteria_match__should_fail_to_match_on_non_matching_birth_range(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("birth_range=2000-2100"))

    def test_criteria_match__should_fail_to_match_when_the_provided_birth_year_range_is_not_numeric(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("birth_range=1900-not_a_number"))

    def test_criteria_match__should_fail_to_match_when_the_provided_birth_year_range_does_not_have_a_separator(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("birth_range=not_a_number"))

    def test_criteria_match__should_match_on_matching_death_year(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertTrue(individual.criteria_match("death=1999"))

    def test_criteria_match__should_fail_to_match_on_matching_death_year(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("death=2127"))

    def test_criteria_match__should_fail_to_match_when_the_provided_death_year_is_not_numeric(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("death=not_a_number"))

    def test_criteria_match__should_match_on_matching_death_range(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertTrue(individual.criteria_match("death_range=1900-2000"))

    def test_criteria_match__should_fail_to_match_on_non_matching_death_range(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("death_range=2000-2100"))

    def test_criteria_match__should_fail_to_match_when_the_provided_death_year_range_is_not_numeric(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("death_range=1900-not_a_number"))

    def test_criteria_match__should_fail_to_match_when_the_provided_death_year_range_does_not_have_a_separator(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("death_range=not_a_number"))

    def test_criteria_match__should_match_on_multiple_criteria(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertTrue(individual.criteria_match("name=First:surname=Last"))

    def test_criteria_match__should_not_match_when_single_criterion_is_missing_a_separator(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("nameFirst"))

    def test_criteria_match__should_not_match_when_multiple_criteria_and_one__is_missing_a_separator(self):
        individual = self._parse_use_case_and_get_individual_element(self._criteria_match_use_case, "@I1@")
        self.assertFalse(individual.criteria_match("name=First:surnameLast"))

    # ------------------------------ START OF HELPER METHODS -----------------------

    @staticmethod
    def _convert_gedcom_string_into_parsable_content(gedcom_file_contents_test_string):
        # Ignores whitespace "lines" at the start and end of the string - allows prettier presentation in the tests.
        # Ignores leading and trailing whitespace on each line - allows for indentation of lines to show clearer test strings.
        return [(a.strip() + '\n').encode('utf-8-sig') for a in gedcom_file_contents_test_string.strip().splitlines()]

    @staticmethod
    def _get_element_by_pointer(root_child_elements, pointer):
        for element in root_child_elements:
            if element.get_pointer() == pointer:
                return element

    @classmethod
    def _parse_use_case_and_get_individual_element(cls, use_case, pointer):
        gedcom_parser = Parser()
        gedcom_parser.parse(cls._convert_gedcom_string_into_parsable_content(use_case))
        return cls._get_element_by_pointer(gedcom_parser.get_root_child_elements(), pointer)



"""GEDCOM element consisting of tag `gedcom.tags.GEDCOM_TAG_INDIVIDUAL`"""

import re as regex
from gedcom.element.element import Element
from gedcom.helpers import deprecated
import gedcom.tags


class NotAnActualIndividualError(Exception):
    pass


class IndividualElement(Element):

    def get_tag(self):
        return gedcom.tags.GEDCOM_TAG_INDIVIDUAL

    def is_deceased(self):
        """Checks if this individual is deceased
        :rtype: bool
        """
        return self._is_tag_present(gedcom.tags.GEDCOM_TAG_DEATH)

    def is_child(self):
        """Checks if this element is a child of a family
        :rtype: bool
        """
        return self._is_tag_present(gedcom.tags.GEDCOM_TAG_FAMILY_CHILD)

    def is_spouse(self):
        """Checks if this element is a spouse in a family
        :rtype: bool
        """
        return self._is_tag_present(gedcom.tags.GEDCOM_TAG_FAMILY_SPOUSE)

    def is_private(self):
        """Checks if this individual is marked private
        :rtype: bool
        """
        for child in self.get_child_elements():
            if child.get_tag() == gedcom.tags.GEDCOM_TAG_PRIVATE:
                private = child.get_value()
                if private == 'Y':
                    return True

        return False

    def get_name(self):
        """Returns an individual's names as a tuple: (`str` given_name, `str` surname)
        If this person has a suffix (e.g., 'Jr') it gets removed entirely.
        :rtype: tuple
        """
        given_name = ""
        surname = ""

        # Return the first gedcom.tags.GEDCOM_TAG_NAME that is found.
        # Alternatively as soon as we have both the gedcom.tags.GEDCOM_TAG_GIVEN_NAME and _SURNAME return those.
        found_given_name = False
        found_surname_name = False

        for child in self.get_child_elements():
            if child.get_tag() == gedcom.tags.GEDCOM_TAG_NAME:
                # Some GEDCOM files don't use child tags but instead
                # place the name in the value of the NAME tag.
                if child.get_value() != "":
                    name = child.get_value().split('/')

                    if len(name) > 0:
                        given_name = name[0].strip()
                        if len(name) > 1:
                            surname = name[1].strip()

                    return given_name, surname

                for childOfChild in child.get_child_elements():

                    if childOfChild.get_tag() == gedcom.tags.GEDCOM_TAG_GIVEN_NAME:
                        given_name = childOfChild.get_value()
                        found_given_name = True

                    if childOfChild.get_tag() == gedcom.tags.GEDCOM_TAG_SURNAME:
                        surname = childOfChild.get_value()
                        found_surname_name = True

                if found_given_name and found_surname_name:
                    return given_name, surname

        # If we reach here we are probably returning empty strings
        return given_name, surname

    def get_all_names(self):
        return [a.get_value() for a in self.get_child_elements() if a.get_tag() == gedcom.tags.GEDCOM_TAG_NAME]

    def surname_match(self, surname_to_match):
        """Matches a string with the surname of an individual
        :type surname_to_match: str
        :rtype: bool
        """
        (given_name, surname) = self.get_name()
        return regex.search(surname_to_match, surname, regex.IGNORECASE)

    @deprecated
    def given_match(self, name):
        """Matches a string with the given name of an individual
        ::deprecated:: As of version 1.0.0 use `given_name_match()` method instead
        :type name: str
        :rtype: bool
        """
        return self.given_name_match(name)

    def given_name_match(self, given_name_to_match):
        """Matches a string with the given name of an individual
        :type given_name_to_match: str
        :rtype: bool
        """
        (given_name, surname) = self.get_name()
        return regex.search(given_name_to_match, given_name, regex.IGNORECASE)

    def get_gender(self):
        """Returns the gender of a person in string format
        :rtype: str
        """
        gender = ""

        for child in self.get_child_elements():
            if child.get_tag() == gedcom.tags.GEDCOM_TAG_SEX:
                gender = child.get_value()

        return gender

    def __get_data_for_date_bearing_tag(self, tag):
        date = ""
        place = ""
        sources = []

        for child in self.get_child_elements():
            if child.get_tag() == tag:
                for childOfChild in child.get_child_elements():

                    if childOfChild.get_tag() == gedcom.tags.GEDCOM_TAG_DATE:
                        date = childOfChild.get_value()

                    if childOfChild.get_tag() == gedcom.tags.GEDCOM_TAG_PLACE:
                        place = childOfChild.get_value()

                    if childOfChild.get_tag() == gedcom.tags.GEDCOM_TAG_SOURCE:
                        sources.append(childOfChild.get_value())

        return date, place, sources

    def get_birth_data(self):
        """Returns the birth data of a person formatted as a tuple: (`str` date, `str` place, `list` sources)
        :rtype: tuple
        """
        return self.__get_data_for_date_bearing_tag(gedcom.tags.GEDCOM_TAG_BIRTH)

    def get_death_data(self):
        """Returns the death data of a person formatted as a tuple: (`str` date, `str` place, `list` sources)
        :rtype: tuple
        """
        return self.__get_data_for_date_bearing_tag(gedcom.tags.GEDCOM_TAG_DEATH)

    def __get_year_in_tag_type(self, tag):
        date = ""

        for child in self.get_child_elements():
            if child.get_tag() == tag:
                for childOfChild in child.get_child_elements():
                    if childOfChild.get_tag() == gedcom.tags.GEDCOM_TAG_DATE:
                        date_split = childOfChild.get_value().split()
                        date = date_split[len(date_split) - 1]

        if date == "":
            return -1
        try:
            return int(date)
        except ValueError:
            return -1

    def get_birth_year(self):
        """Returns the birth year of a person in integer format
        :rtype: int
        """
        return self.__get_year_in_tag_type(gedcom.tags.GEDCOM_TAG_BIRTH)

    def get_death_year(self):
        """Returns the death year of a person in integer format
        :rtype: int
        """
        return self.__get_year_in_tag_type(gedcom.tags.GEDCOM_TAG_DEATH)

    @deprecated
    def get_burial(self):
        """Returns the burial data of a person formatted as a tuple: (`str` date, `str´ place, `list` sources)
        ::deprecated:: As of version 1.0.0 use `get_burial_data()` method instead
        :rtype: tuple
        """
        self.get_burial_data()

    def get_burial_data(self):
        """Returns the burial data of a person formatted as a tuple: (`str` date, `str´ place, `list` sources)
        :rtype: tuple
        """
        return self.__get_data_for_date_bearing_tag(gedcom.tags.GEDCOM_TAG_BURIAL)

    @deprecated
    def get_census(self):
        """Returns a list of censuses of an individual formatted as tuples: (`str` date, `str´ place, `list` sources)
        ::deprecated:: As of version 1.0.0 use `get_census_data()` method instead
        :rtype: list of tuple
        """
        self.get_census_data()

    def get_census_data(self):
        """Returns a list of censuses of an individual formatted as tuples: (`str` date, `str´ place, `list` sources)
        :rtype: list of tuple
        """
        census = []

        for child in self.get_child_elements():
            if child.get_tag() == gedcom.tags.GEDCOM_TAG_CENSUS:

                date = ''
                place = ''
                sources = []

                for childOfChild in child.get_child_elements():

                    if childOfChild.get_tag() == gedcom.tags.GEDCOM_TAG_DATE:
                        date = childOfChild.get_value()

                    if childOfChild.get_tag() == gedcom.tags.GEDCOM_TAG_PLACE:
                        place = childOfChild.get_value()

                    if childOfChild.get_tag() == gedcom.tags.GEDCOM_TAG_SOURCE:
                        sources.append(childOfChild.get_value())

                census.append((date, place, sources))

        return census

    def get_last_change_date(self):
        """Returns the date of when the person data was last changed formatted as a string
        :rtype: str
        """
        date = ""

        for child in self.get_child_elements():
            if child.get_tag() == gedcom.tags.GEDCOM_TAG_CHANGE:
                for childOfChild in child.get_child_elements():
                    if childOfChild.get_tag() == gedcom.tags.GEDCOM_TAG_DATE:
                        date = childOfChild.get_value()

        return date

    def get_occupation(self):
        """Returns the occupation of a person
        :rtype: str
        """
        occupation = ""

        for child in self.get_child_elements():
            if child.get_tag() == gedcom.tags.GEDCOM_TAG_OCCUPATION:
                occupation = child.get_value()

        return occupation

    def birth_year_match(self, year):
        """Returns `True` if the given year matches the birth year of this person
        :type year: int
        :rtype: bool
        """
        return self.get_birth_year() == year

    def birth_range_match(self, from_year, to_year):
        """Checks if the birth year of a person lies within the given range
        :type from_year: int
        :type to_year: int
        :rtype: bool
        """
        birth_year = self.get_birth_year()

        if from_year <= birth_year <= to_year:
            return True

        return False

    def death_year_match(self, year):
        """Returns `True` if the given year matches the death year of this person
        :type year: int
        :rtype: bool
        """
        return self.get_death_year() == year

    def death_range_match(self, from_year, to_year):
        """Checks if the death year of a person lies within the given range
        :type from_year: int
        :type to_year: int
        :rtype: bool
        """
        death_year = self.get_death_year()

        if from_year <= death_year <= to_year:
            return True

        return False

    def criteria_match(self, criteria):
        """Checks if this individual matches all of the given criteria

        `criteria` is a colon-separated list, where each item in the
        list has the form [name]=[value]. The following criteria are supported:

        surname=[name]
             Match a person with [name] in any part of the `surname`.
        given_name=[given_name]
             Match a person with [given_name] in any part of the given `given_name`.
        birth=[year]
             Match a person whose birth year is a four-digit [year].
        birth_range=[from_year-to_year]
             Match a person whose birth year is in the range of years from
             [from_year] to [to_year], including both [from_year] and [to_year].
        death=[year]
             Match a person whose death year is a four-digit [year].
        death_range=[from_year-to_year]
             Match a person whose death year is in the range of years from
             [from_year] to [to_year], including both [from_year] and [to_year].

        :type criteria: str
        :rtype: bool
        """

        match = True

        # Check if criteria is a valid criteria and can be split by `:` and `=` characters
        if ":" in criteria:
            for criterion in criteria.split(':'):
                if "=" not in criterion:
                    return False
        else:
            if "=" not in criteria:
                return False

        for criterion in criteria.split(':'):
            key, value = criterion.split('=')

            if key == "surname" and not self.surname_match(value):
                match = False
            elif key == "name" and not self.given_name_match(value):
                match = False
            elif key == "birth":

                try:
                    year = int(value)
                    if not self.birth_year_match(year):
                        match = False
                except ValueError:
                    match = False

            elif key == "birth_range":

                try:
                    from_year, to_year = value.split('-')
                    from_year = int(from_year)
                    to_year = int(to_year)
                    if not self.birth_range_match(from_year, to_year):
                        match = False
                except ValueError:
                    match = False

            elif key == "death":

                try:
                    year = int(value)
                    if not self.death_year_match(year):
                        match = False
                except ValueError:
                    match = False

            elif key == "death_range":

                try:
                    from_year, to_year = value.split('-')
                    from_year = int(from_year)
                    to_year = int(to_year)
                    if not self.death_range_match(from_year, to_year):
                        match = False
                except ValueError:
                    match = False

        return match



"""GEDCOM element consisting of tag `gedcom.tags.GEDCOM_TAG_FILE`"""

from gedcom.element.element import Element


class NotAnActualFileError(Exception):
    pass


class FileElement(Element):
    pass



"""GEDCOM element consisting of tag `gedcom.tags.GEDCOM_TAG_FAMILY`"""

from gedcom.element.element import Element
import gedcom.tags


class NotAnActualFamilyError(Exception):
    pass


class FamilyElement(Element):

    def get_tag(self):
        return gedcom.tags.GEDCOM_TAG_FAMILY

    def has_children(self):
        """Returns whether or not there is at least one child in this family.
        :rtype: boolean
        """
        return self._is_tag_present(gedcom.tags.GEDCOM_TAG_CHILD)



import unittest

from gedcom.parser import Parser


class TestFamilyElement(unittest.TestCase):

    def test_has_children__a_family_without_children_has_no_children(self):
        use_case = """
            0 @F1@ FAM
                1 HUSB @I5@
                1 WIFE @I6@
        """
        family = self._parse_use_case_and_get_individual_element(use_case, "@F1@")
        self.assertFalse(family.has_children())

    def test_has_children__a_family_with_at_least_one_child_has_children(self):
        use_case = """
            0 @F1@ FAM
                1 HUSB @I5@
                1 WIFE @I6@
                1 CHIL @I3@
        """
        family = self._parse_use_case_and_get_individual_element(use_case, "@F1@")
        self.assertTrue(family.has_children())

    # ------------------------------ START OF HELPER METHODS -----------------------

    @staticmethod
    def _convert_gedcom_string_into_parsable_content(gedcom_file_contents_test_string):
        # Ignores whitespace "lines" at the start and end of the string - allows prettier presentation in the tests.
        # Ignores leading and trailing whitespace on each line - allows for indentation of lines to show clearer test strings.
        return [(a.strip() + '\n').encode('utf-8-sig') for a in gedcom_file_contents_test_string.strip().splitlines()]

    @staticmethod
    def _get_element_by_pointer(root_child_elements, pointer):
        for element in root_child_elements:
            if element.get_pointer() == pointer:
                return element

    @classmethod
    def _parse_use_case_and_get_individual_element(cls, use_case, pointer):
        gedcom_parser = Parser()
        gedcom_parser.parse(cls._convert_gedcom_string_into_parsable_content(use_case))
        return cls._get_element_by_pointer(gedcom_parser.get_root_child_elements(), pointer)



"""
GEDCOM tags.
"""

GEDCOM_TAG_ABBREVIATION = "ABBR"
"""Value: `ABBR`

A short name of a title, description, or name.
"""

GEDCOM_TAG_ADDRESS = "ADDR"
"""Value: `ADDR`

The contemporary place, usually required for postal purposes, of an individual, a submitter of information, 
a repository, a business, a school, or a company.
"""

GEDCOM_TAG_ADDRESS1 = "ADR1"
"""Value: `ADR1`

The first line of an address.
"""

GEDCOM_TAG_ADDRESS2 = "ADR2"
"""Value: `ADR2`

The second line of an address.
"""

GEDCOM_TAG_ADOPTION = "ADOP"
"""Value: `ADOP`

Pertaining to creation of a child-parent relationship that does not exist biologically.
"""

GEDCOM_TAG_ANCESTRAL_FILE_NUMBER = "AFN"
"""Value: `AFN`

Ancestral File Number, a unique permanent record file number of an individual record stored in Ancestral File.
"""

GEDCOM_TAG_AGE = "AGE"
"""Value: `AGE`

The age of the individual at the time an event occurred, or the age listed in the document.
"""

GEDCOM_TAG_AGENCY = "AGNC"
"""Value: `AGNC`

The institution or individual having authority and/or responsibility to manage or govern.
"""

GEDCOM_TAG_ALIAS = "ALIA"
"""Value: `ALIA`

An indicator to link different record descriptions of a person who may be the same person.
"""

GEDCOM_TAG_ANCESTORS = "ANCE"
"""Value: `ANCE`

Pertaining to forbearers of an individual.
"""

GEDCOM_TAG_ANCES_INTEREST = "ANCI"
"""Value: `ANCI`

Indicates an interest in additional research for ancestors of this individual. (See also DESI)
"""

GEDCOM_TAG_ANNULMENT = "ANUL"
"""Value: `ANUL`

Declaring a marriage void from the beginning (never existed).
"""

GEDCOM_TAG_ASSOCIATES = "ASSO"
"""Value: `ASSO`

An indicator to link friends, neighbors, relatives, or associates of an individual.
"""

GEDCOM_TAG_AUTHOR = "AUTH"
"""Value: `AUTH`

The name of the individual who created or compiled information.
"""

GEDCOM_TAG_BAPTISM_LDS = "BAPL"
"""Value: `BAPL`

The event of baptism performed at age eight or later by priesthood authority of the LDS Church. (See also BAPM)
"""

GEDCOM_TAG_BAPTISM = "BAPM"
"""Value: `BAPM`

The event of baptism (not LDS), performed in infancy or later. (See also BAPL and CHR)
"""

GEDCOM_TAG_BAR_MITZVAH = "BARM"
"""Value: `BARM`

The ceremonial event held when a Jewish boy reaches age 13.
"""

GEDCOM_TAG_BAS_MITZVAH = "BASM"
"""Value: `BASM`

The ceremonial event held when a Jewish girl reaches age 13, also known as "Bat Mitzvah."
"""

GEDCOM_TAG_BIRTH = "BIRT"
"""Value: `BIRT`

The event of entering into life.
"""

GEDCOM_TAG_BLESSING = "BLES"
"""Value: `BLES`

A religious event of bestowing divine care or intercession. Sometimes given in connection with a naming ceremony.
"""

GEDCOM_TAG_BINARY_OBJECT = "BLOB"
"""Value: `BLOB`

A grouping of data used as input to a multimedia system that processes binary data to represent images, sound, and video. deleted in Gedcom 5.5.1
"""

GEDCOM_TAG_BURIAL = "BURI"
"""Value: `BURI`

The event of the proper disposing of the mortal remains of a deceased person.
"""

GEDCOM_TAG_CALL_NUMBER = "CALN"
"""Value: `CALN`

The number used by a repository to identify the specific items in its collections.
"""

GEDCOM_TAG_CASTE = "CAST"
"""Value: `CAST`

The name of an individual's rank or status in society, based on racial or religious differences, or differences in wealth, inherited rank, profession, 
occupation, etc.
"""

GEDCOM_TAG_CAUSE = "CAUS"
"""Value: `CAUS`

A description of the cause of the associated event or fact, such as the cause of death.
"""

GEDCOM_TAG_CENSUS = "CENS"
"""Value: `CENS`.

The event of the periodic count of the population for a designated locality, such as a national or state Census.
"""

GEDCOM_TAG_CHANGE = "CHAN"
"""Value: `CHAN`

Indicates a change, correction, or modification. Typically used in connection
with a `gedcom.tags.GEDCOM_TAG_DATE` to specify when a change in information occurred.
"""

GEDCOM_TAG_CHARACTER = "CHAR"
"""Value: `CHAR`

An indicator of the character set used in writing this automated information.
"""

GEDCOM_TAG_CHILD = "CHIL"
"""Value: `CHIL`

The natural, adopted, or sealed (LDS) child of a father and a mother.
"""

GEDCOM_TAG_CHRISTENING = "CHR"
"""Value: `CHR`

The religious event (not LDS) of baptizing and/or naming a child.
"""

GEDCOM_TAG_ADULT_CHRISTENING = "CHRA"
"""Value: `CHRA`

The religious event (not LDS) of baptizing and/or naming an adult person.
"""

GEDCOM_TAG_CITY = "CITY"
"""Value: `CITY`

A lower level jurisdictional unit. Normally an incorporated municipal unit.
"""

GEDCOM_TAG_CONCATENATION = "CONC"
"""Value: `CONC`

An indicator that additional data belongs to the superior value. The information from the `CONC` value is to
be connected to the value of the superior preceding line without a space and without a carriage return and/or
new line character. Values that are split for a `CONC` tag must always be split at a non-space. If the value is
split on a space the space will be lost when concatenation takes place. This is because of the treatment that
spaces get as a GEDCOM delimiter, many GEDCOM values are trimmed of trailing spaces and some systems look for
the first non-space starting after the tag to determine the beginning of the value.
"""

GEDCOM_TAG_CONFIRMATION = "CONF"
"""Value: `CONF`

The religious event (not LDS) of conferring the gift of the Holy Ghost and, among protestants, full church membership.
"""

GEDCOM_TAG_CONFIRMATION_L = "CONL"
"""Value: `CONL`

The religious event by which a person receives membership in the LDS Church.
"""

GEDCOM_TAG_CONTINUED = "CONT"
"""Value: `CONT`

An indicator that additional data belongs to the superior value. The information from the `CONT` value is to be
connected to the value of the superior preceding line with a carriage return and/or new line character.
Leading spaces could be important to the formatting of the resultant text. When importing values from `CONT` lines
the reader should assume only one delimiter character following the `CONT` tag. Assume that the rest of the leading
spaces are to be a part of the value.
"""

GEDCOM_TAG_COPYRIGHT = "COPR"
"""Value: `COPR`

A statement that accompanies data to protect it from unlawful duplication and distribution.
"""

GEDCOM_TAG_CORPORATE = "CORP"
"""Value: `CORP`

A name of an institution, agency, corporation, or company.
"""

GEDCOM_TAG_CREMATION = "CREM"
"""Value: `CREM`

Disposal of the remains of a person's body by fire.
"""

GEDCOM_TAG_COUNTRY = "CTRY"
"""Value: `CTRY`

The name or code of the country.
"""

GEDCOM_TAG_DATA = "DATA"
"""Value: `DATA`

Pertaining to stored automated information.
"""

GEDCOM_TAG_DATE = "DATE"
"""Value: `DATE`

The time of an event in a calendar format.
"""

GEDCOM_TAG_DEATH = "DEAT"
"""Value: `DEAT`

The event when mortal life terminates.
"""

GEDCOM_TAG_DESCENDANTS = "DESC"
"""Value: `DESC`

Pertaining to offspring of an individual.
"""

GEDCOM_TAG_DESCENDANT_INT = "DESI"
"""Value: `DESI`

Indicates an interest in research to identify additional descendants of this individual. (See also ANCI)
"""

GEDCOM_TAG_DESTINATION = "DEST"
"""Value: `DEST`

A system receiving data.
"""

GEDCOM_TAG_DIVORCE = "DIV"
"""Value: `DIV`

An event of dissolving a marriage through civil action.
"""

GEDCOM_TAG_DIVORCE_FILED = "DIVF"
"""Value: `DIVF`

An event of filing for a divorce by a spouse.
"""

GEDCOM_TAG_PHY_DESCRIPTION = "DSCR"
"""Value: `DSCR`

The physical characteristics of a person, place, or thing.
"""

GEDCOM_TAG_EDUCATION = "EDUC"
"""Value: `EDUC`

Indicator of a level of education attained.
"""

GEDCOM_TAG_EMAIL = "EMAIL"
"""Value: `EMAIL`

An electronic address that can be used for contact such as an email address... new in Gedcom 5.5.1
"""

GEDCOM_TAG_EMIGRATION = "EMIG"
"""Value: `EMIG`

An event of leaving one's homeland with the intent of residing elsewhere.
"""

GEDCOM_TAG_ENDOWMENT = "ENDL"
"""Value: `ENDL`

A religious event where an endowment ordinance for an individual was performed by priesthood authority in an LDS temple.
"""

GEDCOM_TAG_ENGAGEMENT = "ENGA"
"""Value: `ENGA`

An event of recording or announcing an agreement between two people to become married.
"""

GEDCOM_TAG_EVENT = "EVEN"
"""Value: `EVEN`

A noteworthy happening related to an individual, a group, or an organization.
"""

GEDCOM_TAG_FACT = "FACT"
"""Value: `FACT`

Pertaining to a noteworthy attribute or fact concerning an individual, a group, or an organization. A FACT structure is usually qualified or classified by 
a subordinate use of the TYPE tag. new in Gedcom 5.5.1
"""

GEDCOM_TAG_FAMILY = "FAM"
"""Value: `FAM`.

Identifies a legal, common law, or other customary relationship of man and woman and their children,
if any, or a family created by virtue of the birth of a child to its biological father and mother.
"""

GEDCOM_TAG_FAMILY_CHILD = "FAMC"
"""Value: `FAMC`

Identifies the family in which an individual appears as a child.
"""

GEDCOM_TAG_FAMILY_FILE = "FAMF"
"""Value: `FAMF`

Pertaining to, or the name of, a family file. Names stored in a file that are assigned to a family for doing temple ordinance work.
"""

GEDCOM_TAG_FAMILY_SPOUSE = "FAMS"
"""Value: `FAMS`

Identifies the family in which an individual appears as a spouse.
"""

GEDCOM_TAG_FAX = "FAX"
"""Value: `FAX`

A FAX telephone number appropriate for sending data facsimiles. new in Gedcom 5.5.1
"""

GEDCOM_TAG_FIRST_COMMUNION = "FCOM"
"""Value: `FCOM`

A religious rite, the first act of sharing in the Lord's supper as part of church worship.
"""

GEDCOM_TAG_FILE = "FILE"
"""Value: `FILE`

An information storage place that is ordered and arranged for preservation and reference.
"""

GEDCOM_TAG_PHONETIC = "FONE"
"""Value: `FONE`

A phonetic variation of a superior text string. new in Gedcom 5.5.1
"""

GEDCOM_TAG_FORMAT = "FORM"
"""Value: `FORM`

An assigned name given to a consistent format in which information can be conveyed.
"""

GEDCOM_PROGRAM_DEFINED_TAG_FREL = "_FREL"
"""Value: `_FREL`

Relationship to a father.
"""

GEDCOM_TAG_GEDCOM = "GEDC"
"""Value: `GEDC`

Information about the use of GEDCOM in a transmission.
"""

GEDCOM_TAG_GIVEN_NAME = "GIVN"
"""Value: `GIVN`

A given or earned name used for official identification of a person.
"""

GEDCOM_TAG_GRADUATION = "GRAD"
"""Value: `GRAD`

An event of awarding educational diplomas or degrees to individuals.
"""

GEDCOM_TAG_HEADER = "HEAD"
"""Value: `HEAD`

Identifies information pertaining to an entire GEDCOM transmission.
"""

GEDCOM_TAG_HUSBAND = "HUSB"
"""Value: `HUSB`

An individual in the family role of a married man or father.
"""

GEDCOM_TAG_IDENT_NUMBER = "IDNO"
"""Value: `IDNO`

A number assigned to identify a person within some significant external system.
"""

GEDCOM_TAG_IMMIGRATION = "IMMI"
"""Value: `IMMI`

An event of entering into a new locality with the intent of residing there.
"""

GEDCOM_TAG_INDIVIDUAL = "INDI"
"""Value: `INDI`

A person.
"""

GEDCOM_TAG_LANGUAGE = "LANG"
"""Value: `LANG`

The name of the language used in a communication or transmission of information.
"""

GEDCOM_TAG_LATITUDE = "LATI"
"""Value: `LATI`

A value indicating a coordinate position on a line, plane, or space. new in Gedcom 5.5.1
"""

GEDCOM_TAG_LEGATEE = "LEGA"
"""Value: `LEGA`

A role of an individual acting as a person receiving a bequest or legal devise.
"""

GEDCOM_TAG_LONGITUDE = "LONG"
"""Value: `LONG`

A value indicating a coordinate position on a line, plane, or space. new in Gedcom 5.5.1
"""

GEDCOM_TAG_MAP = "MAP"
"""Value: `MAP`

Pertains to a representation of measurements usually presented in a graphical form. new in Gedcom 5.5.1
"""

GEDCOM_TAG_MARRIAGE_BANN = "MARB"
"""Value: `MARB`

An event of an official public notice given that two people intend to marry.
"""

GEDCOM_TAG_MARR_CONTRACT = "MARC"
"""Value: `MARC`

An event of recording a formal agreement of marriage, including the prenuptial agreement in which marriage partners reach 
agreement about the property rights of one or both, securing property to their children.
"""

GEDCOM_TAG_MARR_LICENSE = "MARL"
"""Value: `MARL`

An event of obtaining a legal license to marry.
"""

GEDCOM_TAG_MARRIAGE = "MARR"
"""Value: `MARR`.

A legal, common-law, or customary event of creating a family unit of a man and a woman as husband and wife.
"""

GEDCOM_TAG_MARR_SETTLEMENT = "MARS"
"""Value: `MARS`

An event of creating an agreement between two people contemplating marriage, at which time they agree to release or modify 
property rights that would otherwise arise from the marriage.
"""

GEDCOM_TAG_MEDIA = "MEDI"
"""Value: `MEDI`

Identifies information about the media or having to do with the medium in which information is stored.
"""

GEDCOM_PROGRAM_DEFINED_TAG_MREL = "_MREL"
"""Value: `_MREL`

Relationship to a mother.
"""

GEDCOM_TAG_NAME = "NAME"
"""Value: `NAME`.

A word or combination of words used to help identify an individual, title, or other item.
More than one NAME line should be used for people who were known by multiple names.
"""

GEDCOM_TAG_NATIONALITY = "NATI"
"""Value: `NATI`

The national heritage of an individual.
"""

GEDCOM_TAG_NATURALIZATION = "NATU"
"""Value: `NATU`

The event of obtaining citizenship.
"""

GEDCOM_TAG_CHILDREN_COUNT = "NCHI"
"""Value: `NCHI`

The number of children that this person is known to be the parent of (all marriages) when subordinate to an individual, 
or that belong to this family when subordinate to a FAM_RECORD.
"""

GEDCOM_TAG_NICKNAME = "NICK"
"""Value: `NICK`

A descriptive or familiar that is used instead of, or in addition to, one's proper name.
"""

GEDCOM_TAG_MARRIAGE_COUNT = "NMR"
"""Value: `NMR`

The number of times this person has participated in a family as a spouse or parent.
"""

GEDCOM_TAG_NOTE = "NOTE"
"""Value: `NOTE`

Additional information provided by the submitter for understanding the enclosing data.
"""

GEDCOM_TAG_NAME_PREFIX = "NPFX"
"""Value: `NPFX`

Text which appears on a name line before the given and surname parts of a name. i.e. ( Lt. Cmndr. ) Joseph /Allen/ jr. 
In this example Lt. Cmndr. is considered as the name prefix portion.
"""

GEDCOM_TAG_NAME_SUFFIX = "NSFX"
"""Value: `NSFX`

Text which appears on a name line after or behind the given and surname parts of a name. i.e. Lt. Cmndr. Joseph /Allen/ ( jr. ) 
In this example jr. is considered as the name suffix portion.
"""

GEDCOM_TAG_OBJECT = "OBJE"
"""Value: `OBJE`

Pertaining to a grouping of attributes used in describing something. Usually referring to the data required
to represent a multimedia object, such an audio recording, a photograph of a person, or an image of a document.
"""

GEDCOM_TAG_OCCUPATION = "OCCU"
"""Value: `OCCU`

The type of work or profession of an individual.
"""

GEDCOM_TAG_ORDINANCE = "ORDI"
"""Value: `ORDI`

Pertaining to a religious ordinance in general.
"""

GEDCOM_TAG_ORDINATION = "ORDN"
"""Value: `ORDN`

A religious event of receiving authority to act in religious matters.
"""

GEDCOM_TAG_PAGE = "PAGE"
"""Value: `PAGE`

A number or description to identify where information can be found in a referenced work.
"""

GEDCOM_TAG_PEDIGREE = "PEDI"
"""Value: `PEDI`

Information pertaining to an individual to parent lineage chart.
"""

GEDCOM_TAG_PHONE = "PHON"
"""Value: `PHON`

A unique number assigned to access a specific telephone.
"""

GEDCOM_TAG_PLACE = "PLAC"
"""Value: `PLAC`

A jurisdictional name to identify the place or location of an event.
"""

GEDCOM_TAG_POSTAL_CODE = "POST"
"""Value: `POST`

A code used by a postal service to identify an area to facilitate mail handling.
"""

GEDCOM_TAG_PRIVATE = "PRIV"
"""Value: `PRIV`

Flag for private address or event.
"""

GEDCOM_TAG_PROBATE = "PROB"
"""Value: `PROB`

An event of judicial determination of the validity of a will. May indicate several related court activities over several dates.
"""

GEDCOM_TAG_PROPERTY = "PROP"
"""Value: `PROP`

Pertaining to possessions such as real estate or other property of interest.
"""

GEDCOM_TAG_PUBLICATION = "PUBL"
"""Value: `PUBL`

Refers to when and/or were a work was published or created.
"""

GEDCOM_TAG_QUALITY_OF_DATA = "QUAY"
"""Value: `QUAY`

An assessment of the certainty of the evidence to support the conclusion drawn from evidence.
"""

GEDCOM_TAG_REFERENCE = "REFN"
"""Value: `REFN`

A description or number used to identify an item for filing, storage, or other reference purposes.
"""

GEDCOM_TAG_RELATIONSHIP = "RELA"
"""Value: `RELA`

A relationship value between the indicated contexts.
"""

GEDCOM_TAG_RELIGION = "RELI"
"""Value: `RELI`

A religious denomination to which a person is affiliated or for which a record applies.
"""

GEDCOM_TAG_REPOSITORY = "REPO"
"""Value: `REPO`

An institution or person that has the specified item as part of their collection(s).
"""

GEDCOM_TAG_RESIDENCE = "RESI"
"""Value: `RESI`

The act of dwelling at an address for a period of time.
"""

GEDCOM_TAG_RESTRICTION = "RESN"
"""Value: `RESN`

A processing indicator signifying access to information has been denied or otherwise restricted.
"""

GEDCOM_TAG_RETIREMENT = "RETI"
"""Value: `RETI`

An event of exiting an occupational relationship with an employer after a qualifying time period.
"""

GEDCOM_TAG_REC_FILE_NUMBER = "RFN"
"""Value: `RFN`

A permanent number assigned to a record that uniquely identifies it within a known file.
"""

GEDCOM_TAG_REC_ID_NUMBER = "RIN"
"""Value: `RIN`

A number assigned to a record by an originating automated system that can be used by a receiving system to report results pertaining to that record.
"""

GEDCOM_TAG_ROLE = "ROLE"
"""Value: `ROLE`

A name given to a role played by an individual in connection with an event.
"""

GEDCOM_TAG_ROMANIZED = "ROMN"
"""Value: `ROMN`

A romanized variation of a superior text string. new in Gedcom 5.5.1
"""

GEDCOM_TAG_SEX = "SEX"
"""Value: `SEX`

Indicates the sex of an individual--male or female.
"""

GEDCOM_TAG_SEALING_CHILD = "SLGC"
"""Value: `SLGC`

A religious event pertaining to the sealing of a child to his or her parents in an LDS temple ceremony.
"""

GEDCOM_TAG_SEALING_SPOUSE = "SLGS"
"""Value: `SLGS`

A religious event pertaining to the sealing of a husband and wife in an LDS temple ceremony.
"""

GEDCOM_TAG_SOURCE = "SOUR"
"""Value: `SOUR`

The initial or original material from which information was obtained.
"""

GEDCOM_TAG_SURN_PREFIX = "SPFX"
"""Value: `SPFX`

A name piece used as a non-indexing pre-part of a surname.
"""

GEDCOM_TAG_SOC_SEC_NUMBER = "SSN"
"""Value: `SSN`

A number assigned by the United States Social Security Administration. Used for tax identification purposes.
"""

GEDCOM_TAG_STATE = "STAE"
"""Value: `STAE`

A geographical division of a larger jurisdictional area, such as a State within the United States of America.
"""

GEDCOM_TAG_STATUS = "STAT"
"""Value: `STAT`

An assessment of the state or condition of something.
"""

GEDCOM_TAG_SUBMITTER = "SUBM"
"""Value: `SUBM`

An individual or organization who contributes genealogical data to a file or transfers it to someone else.
"""

GEDCOM_TAG_SUBMISSION = "SUBN"
"""Value: `SUBN`

Pertains to a collection of data issued for processing.
"""

GEDCOM_TAG_SURNAME = "SURN"
"""Value: `SURN`

A family name passed on or used by members of a family.
"""

GEDCOM_TAG_TEMPLE = "TEMP"
"""Value: `TEMP`

The name or code that represents the name a temple of the LDS Church.
"""

GEDCOM_TAG_TEXT = "TEXT"
"""Value: `TEXT`

The exact wording found in an original source document.
"""

GEDCOM_TAG_TIME = "TIME"
"""Value: `TIME`

A time value in a 24-hour clock format, including hours, minutes, and optional seconds, separated by a 
colon (:). Fractions of seconds are shown in decimal notation.
"""

GEDCOM_TAG_TITLE = "TITL"
"""Value: `TITL`

A description of a specific writing or other work, such as the title of a book when used in a source context, 
or a formal designation used by an individual in connection with positions of royalty or other social status, 
such as Grand Duke.
"""

GEDCOM_TAG_TRAILER = "TRLR"
"""Value: `TRLR`

At level 0, specifies the end of a GEDCOM transmission.
"""

GEDCOM_TAG_TYPE = "TYPE"
"""Value: `TYPE`

A further qualification to the meaning of the associated superior tag. The value does not have any computer 
processing reliability. It is more in the form of a short one or two word note that should be displayed any 
time the associated data is displayed.
"""

GEDCOM_TAG_VERSION = "VERS"
"""Value: `VERS`

Indicates which version of a product, item, or publication is being used or referenced.
"""

GEDCOM_TAG_WIFE = "WIFE"
"""Value: `WIFE`

An individual in the role as a mother and/or married woman.
"""

GEDCOM_TAG_WEB = "WWW"
"""Value: `WWW`

World Wide Web home page. new in Gedcom 5.5.1
"""

GEDCOM_TAG_WILL = "WILL"
"""Value: `WILL`

A legal document treated as an event, by which a person disposes of his or her estate, to take effect after 
death. The event date is the date the will was signed while the person was alive. (See also PROBate)
"""



"""Virtual GEDCOM root element containing all logical records as children"""

from gedcom.element.element import Element


class RootElement(Element):
    """Virtual GEDCOM root element containing all logical records as children"""

    def __init__(self, level=-1, pointer="", tag="ROOT", value="", crlf="\n", multi_line=True):
        super(RootElement, self).__init__(level, pointer, tag, value, crlf, multi_line)



from setuptools import setup, find_packages
from os import path
from io import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the `README.md` file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='python-gedcom-2',
    version='1.3.0',
    description='A Python module for parsing, analyzing, and manipulating GEDCOM files.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/swanny0819/python-gedcom',
    author='Patrick Swanson',
    author_email='patswanson+python_gedcom@gmail.com',
    license='GPLv2',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Sociology :: Genealogy',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='python gedcom parser',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=[],
    extras_require={
        'dev': ['setuptools', 'wheel', 'twine', 'pdoc3'],
        'test': ['tox'],
    },
    package_data={},
    data_files=[],
    project_urls={
        'Bug Reports': 'https://github.com/swanny0819/python-gedcom/issues',
        'Source': 'https://github.com/swanny0819/python-gedcom',
    },
)



import unittest

from gedcom.element.element import Element

from gedcom.element.object import ObjectElement

from gedcom.element.individual import IndividualElement

from gedcom.element.file import FileElement

from gedcom.element.family import FamilyElement
from gedcom.element.root import RootElement


class TestElement(unittest.TestCase):

    # --------------------- START OF get_multi_line_value TESTING -----------------------

    # TODO: test_get_multi_line_value__should_get_single_line_multiline_value
    # TODO: test_get_multi_line_value__should_get_and_combine_multiline_values
    # TODO: test_get_multi_line_value__should_return_original_value_passed_into_set_multi_line_value

    # --------------------- START OF set_multi_line_value TESTING -----------------------

    def test_set_multi_line_value__should_take_single_line_value_and_put_it_in_value_of_this_element(self):
        root_element = RootElement()
        multiline_value = "This is a test value that is on a single line"
        root_element.set_multi_line_value(multiline_value)

        child_elements = root_element.get_child_elements()
        self.assertEqual(0, len(child_elements))
        self.assertEqual(multiline_value, root_element.get_value())

    def test_set_multi_line_value__should_take_multiline_value_and_create(self):
        root_element = RootElement()
        multiline_value = "This is a test value\n" \
                          "and so is this\n" \
                          "and this one too"
        root_element.set_multi_line_value(multiline_value)

        child_elements = root_element.get_child_elements()
        self.assertEqual(2, len(child_elements))
        self.assertEqual("This is a test value", root_element.get_value())

        self.assertEqual("and so is this", child_elements[0].get_value())
        self.assertEqual(0, child_elements[0].get_level())
        self.assertEqual("CONT", child_elements[0].get_tag())
        self.assertEqual("", child_elements[0].get_pointer())

        self.assertEqual("and this one too", child_elements[1].get_value())
        self.assertEqual(0, child_elements[1].get_level())
        self.assertEqual("CONT", child_elements[1].get_tag())
        self.assertEqual("", child_elements[1].get_pointer())

    def test_set_multi_line_value__should_remove_existing_value_and_concatenations_but_not_other_nodes(self):
        root_element = RootElement(value="original value")
        root_element.new_child_element("CONC", value="original concatenation")
        root_element.new_child_element("CONT", value="original continued")
        root_element.new_child_element("BIRT", value="1 JUN 2019")

        multiline_value = "This is a test value that is on a single line"
        root_element.set_multi_line_value(multiline_value)

        child_elements = root_element.get_child_elements()
        self.assertEqual(multiline_value, root_element.get_value())
        self.assertEqual(1, len(child_elements))
        self.assertEqual("1 JUN 2019", child_elements[0].get_value())

    # --------------------- START OF new_child_element TESTING -----------------------

    def test_new_child_element__should_create_a_child_family_element(self):
        root_element = RootElement()
        child_element = root_element.new_child_element("FAM", "@F1@")
        self.assertTrue(isinstance(child_element, FamilyElement), child_element)
        self.assertEqual(0, child_element.get_level())
        self.assertEqual("@F1@", child_element.get_pointer())
        self.assertEqual("", child_element.get_value())

    def test_new_child_element__should_create_a_child_file_element(self):
        root_element = RootElement()
        child_element = root_element.new_child_element("FILE", "@F1@")
        self.assertTrue(isinstance(child_element, FileElement), child_element)
        self.assertEqual(0, child_element.get_level())
        self.assertEqual("@F1@", child_element.get_pointer())
        self.assertEqual("", child_element.get_value())

    def test_new_child_element__should_create_a_child_individual_element(self):
        root_element = RootElement()
        child_element = root_element.new_child_element("INDI", "@I1@")
        self.assertTrue(isinstance(child_element, IndividualElement), child_element)
        self.assertEqual(0, child_element.get_level())
        self.assertEqual("@I1@", child_element.get_pointer())
        self.assertEqual("", child_element.get_value())

    def test_new_child_element__should_create_a_child_object_element(self):
        root_element = RootElement()
        child_element = root_element.new_child_element("OBJE", "@O1@")
        self.assertTrue(isinstance(child_element, ObjectElement), child_element)
        self.assertEqual(0, child_element.get_level())
        self.assertEqual("@O1@", child_element.get_pointer())
        self.assertEqual("", child_element.get_value())

    # --------------------- START OF to_gedcom_string TESTING -----------------------

    def test_to_gedcom_string__should_combine_only_the_mandatory_values_non_recursively_for_an_indi_nodes_values(self):
        element = Element(level=0, pointer="@I1@", tag="INDI", value="")
        self.assertEqual("0 @I1@ INDI\n", element.to_gedcom_string())

    def test_to_gedcom_string__should_combine_only_the_mandatory_values_non_recursively_for_a_date_nodes_values(self):
        element = Element(level=2, pointer="", tag="DATE", value="JUN 1 1990")
        self.assertEqual("2 DATE JUN 1 1990\n", element.to_gedcom_string())

    def test_to_gedcom_string__should_combine_only_the_mandatory_values_recursively_for_an_indi_nodes_values(self):
        element = Element(level=0, pointer="@I1@", tag="INDI", value="")
        element.new_child_element("NAME", value="First /Last/")
        birth_element = element.new_child_element("BIRT")
        birth_element.new_child_element("DATE", value="JUN 2 1876")
        expected_string = \
            "0 @I1@ INDI\n" \
            "1 NAME First /Last/\n" \
            "1 BIRT\n" \
            "2 DATE JUN 2 1876\n"
        self.assertEqual(expected_string, element.to_gedcom_string(recursive=True))

    # TODO: test_to_gedcom_string__should_
    # TODO: test_to_gedcom_string__should_
    # TODO: test_to_gedcom_string__should_
    # TODO: test_to_gedcom_string__should_



"""
Base GEDCOM element
"""

from sys import version_info
from gedcom.helpers import deprecated
import gedcom.tags


class Element(object):
    """GEDCOM element

    Each line in a GEDCOM file is an element with the format

    `level [pointer] tag [value]`

    where `level` and `tag` are required, and `pointer` and `value` are
    optional.  Elements are arranged hierarchically according to their
    level, and elements with a level of zero are at the top level.
    Elements with a level greater than zero are children of their
    parent.

    A pointer has the format `@pname@`, where `pname` is any sequence of
    characters and numbers. The pointer identifies the object being
    pointed to, so that any pointer included as the value of any
    element points back to the original object.  For example, an
    element may have a `FAMS` tag whose value is `@F1@`, meaning that this
    element points to the family record in which the associated person
    is a spouse. Likewise, an element with a tag of `FAMC` has a value
    that points to a family record in which the associated person is a
    child.

    See a GEDCOM file for examples of tags and their values.

    Tags available to an element are seen here: `gedcom.tags`
    """

    def __init__(self, level, pointer, tag, value, crlf="\n", multi_line=True):
        # basic element info
        self.__level = level
        self.__pointer = pointer
        self.__tag = tag
        self.__value = value
        self.__crlf = crlf

        # structuring
        self.__children = []
        self.__parent = None

        if multi_line:
            self.set_multi_line_value(value)

    def get_level(self):
        """Returns the level of this element from within the GEDCOM file
        :rtype: int
        """
        return self.__level

    def get_pointer(self):
        """Returns the pointer of this element from within the GEDCOM file
        :rtype: str
        """
        return self.__pointer

    def get_tag(self):
        """Returns the tag of this element from within the GEDCOM file
        :rtype: str
        """
        return self.__tag

    def get_value(self):
        """Return the value of this element from within the GEDCOM file
        :rtype: str
        """
        return self.__value

    def set_value(self, value):
        """Sets the value of this element
        :type value: str
        """
        self.__value = value

    def get_multi_line_value(self):
        """Returns the value of this element including concatenations or continuations
        :rtype: str
        """
        result = self.get_value()
        last_crlf = self.__crlf
        for element in self.get_child_elements():
            tag = element.get_tag()
            if tag == gedcom.tags.GEDCOM_TAG_CONCATENATION:
                result += element.get_value()
                last_crlf = element.__crlf
            elif tag == gedcom.tags.GEDCOM_TAG_CONTINUED:
                result += last_crlf + element.get_value()
                last_crlf = element.__crlf
        return result

    def __available_characters(self):
        """Get the number of available characters of the elements original string
        :rtype: int
        """
        element_characters = len(self.to_gedcom_string())
        return 0 if element_characters > 255 else 255 - element_characters

    def __line_length(self, line):
        """@TODO Write docs.
        :type line: str
        :rtype: int
        """
        total_characters = len(line)
        available_characters = self.__available_characters()
        if total_characters <= available_characters:
            return total_characters
        spaces = 0
        while spaces < available_characters and line[available_characters - spaces - 1] == ' ':
            spaces += 1
        if spaces == available_characters:
            return available_characters
        return available_characters - spaces

    def __set_bounded_value(self, value):
        """@TODO Write docs.
        :type value: str
        :rtype: int
        """
        line_length = self.__line_length(value)
        self.set_value(value[:line_length])
        return line_length

    def __add_bounded_child(self, tag, value):
        """@TODO Write docs.
        :type tag: str
        :type value: str
        :rtype: int
        """
        child = self.new_child_element(tag)
        return child.__set_bounded_value(value)

    def __add_concatenation(self, string):
        """@TODO Write docs.
        :rtype: str
        """
        index = 0
        size = len(string)
        while index < size:
            index += self.__add_bounded_child(gedcom.tags.GEDCOM_TAG_CONCATENATION, string[index:])

    def set_multi_line_value(self, value):
        """Sets the value of this element, adding concatenation and continuation lines when necessary
        :type value: str
        """
        self.set_value('')
        self.__children = [child for child in self.get_child_elements() if
                           child.get_tag() not in (gedcom.tags.GEDCOM_TAG_CONCATENATION, gedcom.tags.GEDCOM_TAG_CONTINUED)]

        lines = value.splitlines()
        if lines:
            line = lines.pop(0)
            n = self.__set_bounded_value(line)
            self.__add_concatenation(line[n:])

            for line in lines:
                n = self.__add_bounded_child(gedcom.tags.GEDCOM_TAG_CONTINUED, line)
                self.__add_concatenation(line[n:])

    def get_child_elements(self):
        """Returns the direct child elements of this element
        :rtype: list of Element
        """
        return self.__children

    def new_child_element(self, tag, pointer="", value=""):
        """Creates and returns a new child element of this element

        :type tag: str
        :type pointer: str
        :type value: str
        :rtype: Element
        """
        from gedcom.element.family import FamilyElement
        from gedcom.element.file import FileElement
        from gedcom.element.individual import IndividualElement
        from gedcom.element.object import ObjectElement

        tag_element_dict = {
            gedcom.tags.GEDCOM_TAG_FAMILY: FamilyElement,
            gedcom.tags.GEDCOM_TAG_FILE: FileElement,
            gedcom.tags.GEDCOM_TAG_INDIVIDUAL: IndividualElement,
            gedcom.tags.GEDCOM_TAG_OBJECT: ObjectElement,
        }

        if tag in tag_element_dict:
            child_element = tag_element_dict[tag](self.get_level() + 1, pointer, tag, value, self.__crlf)
        else:
            child_element = Element(self.get_level() + 1, pointer, tag, value, self.__crlf)

        self.add_child_element(child_element)

        return child_element

    def add_child_element(self, element):
        """Adds a child element to this element

        :type element: Element
        """
        self.get_child_elements().append(element)
        element.set_parent_element(self)

        return element

    def get_parent_element(self):
        """Returns the parent element of this element
        :rtype: Element
        """
        return self.__parent

    def set_parent_element(self, element):
        """Adds a parent element to this element

        There's usually no need to call this method manually,
        `add_child_element()` calls it automatically.

        :type element: Element
        """
        self.__parent = element

    def _is_tag_present(self, tag):
        for child in self.get_child_elements():
            if child.get_tag() == tag:
                return True
        return False

    @deprecated
    def get_individual(self):
        """Returns this element and all of its sub-elements represented as a GEDCOM string
        ::deprecated:: As of version 1.0.0 use `to_gedcom_string()` method instead
        :rtype: str
        """
        return self.to_gedcom_string(True)

    def to_gedcom_string(self, recursive=False):
        """Formats this element and optionally all of its sub-elements into a GEDCOM string
        :type recursive: bool
        :rtype: str
        """

        result = str(self.get_level())

        if self.get_pointer() != "":
            result += ' ' + self.get_pointer()

        result += ' ' + self.get_tag()

        if self.get_value() != "":
            result += ' ' + self.get_value()

        result += self.__crlf

        if self.get_level() < 0:
            result = ''

        if recursive:
            for child_element in self.get_child_elements():
                result += child_element.to_gedcom_string(True)

        return result

    def __str__(self):
        """:rtype: str"""
        if version_info[0] >= 3:
            return self.to_gedcom_string()

        return self.to_gedcom_string().encode('utf-8-sig')



"""GEDCOM element consisting of tag `gedcom.tags.GEDCOM_TAG_OBJECT`"""

from gedcom.element.element import Element
import gedcom.tags


class NotAnActualObjectError(Exception):
    pass


class ObjectElement(Element):

    def is_object(self):
        """Checks if this element is an actual object
        :rtype: bool
        """
        return self.get_tag() == gedcom.tags.GEDCOM_TAG_OBJECT
